﻿using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("GQAP Evaluation", "Class that contains the results of evaluating a solution to the GQAP.")]
  [StorableClass]
  public class Evaluation : Item {

    /// <summary>
    /// The solution is feasible if there is no excess demand
    /// </summary>
    public bool IsFeasible { get { return ExcessDemand <= double.Epsilon; } }

    /// <summary>
    /// The amount of demand that exceeds the respective capacities
    /// </summary>
    [Storable]
    public double ExcessDemand { get; private set; }
    /// <summary>
    /// The amount of slack at each location, is negative in case of excess demand.
    /// </summary>
    /// <remarks>
    /// Slack.Select(x => x &lt; 0 ? -x : 0).Sum() is equivalent to <see cref="ExcessDemand"/>.
    /// </remarks>
    [Storable]
    private double[] slack;
    public IReadOnlyList<double> Slack {
      get { return slack; }
    }
    /// <summary>
    /// The quadratic part of the fitness function that represents the flow or transportation costs.
    /// </summary>
    [Storable]
    public double FlowCosts { get; private set; }
    /// <summary>
    /// The linear part of the fitness function that represents the installation costs.
    /// </summary>
    [Storable]
    public double InstallationCosts { get; private set; }
    
    [StorableConstructor]
    protected Evaluation(bool deserializing) : base(deserializing) { }
    protected Evaluation(Evaluation original, Cloner cloner)
      : base(original, cloner) {
      ExcessDemand = original.ExcessDemand;
      if (original.slack != null)
        slack = original.slack.ToArray();
      FlowCosts = original.FlowCosts;
      InstallationCosts = original.InstallationCosts;
    }
    public Evaluation(double flowCosts, double installationCosts, double excessDemand, double[] slack) {
      FlowCosts = flowCosts;
      InstallationCosts = installationCosts;
      ExcessDemand = excessDemand;
      this.slack = slack;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new Evaluation(this, cloner);
    }
  }
}
