﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("GQAPPathRelinking", "Operator that performs path relinking between two solutions. It is described in Mateus, G., Resende, M., and Silva, R. 2011. GRASP with path-relinking for the generalized quadratic assignment problem. Journal of Heuristics 17, Springer Netherlands, pp. 527-565.")]
  [StorableClass]
  public class GQAPPathRelinking : GQAPCrossover, IQualitiesAwareGQAPOperator {

    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public IScopeTreeLookupParameter<Evaluation> EvaluationParameter {
      get { return (IScopeTreeLookupParameter<Evaluation>)Parameters["Evaluation"]; }
    }

    public IValueParameter<PercentValue> CandidateSizeFactorParameter {
      get { return (IValueParameter<PercentValue>)Parameters["CandidateSizeFactor"]; }
    }

    [StorableConstructor]
    protected GQAPPathRelinking(bool deserializing) : base(deserializing) { }
    protected GQAPPathRelinking(GQAPPathRelinking original, Cloner cloner) : base(original, cloner) { }
    public GQAPPathRelinking()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", ""));
      Parameters.Add(new ScopeTreeLookupParameter<Evaluation>("Evaluation", GQAP.EvaluationDescription));
      Parameters.Add(new ValueParameter<PercentValue>("CandidateSizeFactor", "(η) Determines the size of the set of feasible moves in each path-relinking step relative to the maximum size. A value of 50% means that only half of all possible moves are considered each step.", new PercentValue(0.5)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GQAPPathRelinking(this, cloner);
    }

    public static IntegerVector Apply(IRandom random, ItemArray<IntegerVector> parents, ItemArray<DoubleValue> qualities,
      GQAPInstance problemInstance, double candidateSizeFactor) {
      if (random == null) throw new ArgumentNullException("random", "No IRandom provider is given.");
      if (parents == null || !parents.Any()) throw new ArgumentException("No parents given for path relinking.", "parents");
      if (parents.Length != 2) throw new ArgumentException(String.Format("Two parents were expected for path relinking, but {0} was/were given.", parents.Length.ToString()), "parents");
      if (parents[0].Length != parents[1].Length) throw new ArgumentException("The length of the parents is not equal.", "parents");
      if (qualities == null || qualities.Length == 0) throw new ArgumentException("The qualities are not given.", "qualities");
      if (qualities.Length != parents.Length) throw new ArgumentException(String.Format("There are a different number of parents ({0}) than quality values ({1})", parents.Length.ToString(), qualities.Length.ToString()));
      
      var source = parents[0];
      var target = parents[1];
      IntegerVector result;
      double resultQuality;

      int betterParent = (qualities[0].Value <= qualities[1].Value) ? 0 : 1;

      result = (IntegerVector)parents[betterParent].Clone();
      resultQuality = qualities[betterParent].Value;

      var demands = problemInstance.Demands;
      var capacities = problemInstance.Capacities;

      var cmp = new IntegerVectorEqualityComparer();

      var pi_prime = (IntegerVector)source.Clone();
      var fix = new HashSet<int>();
      var nonFix = new HashSet<int>(Enumerable.Range(0, demands.Length));
      var phi = new HashSet<int>((IntegerVectorEqualityComparer.GetDifferingIndices(pi_prime, target)));

      while (phi.Any()) {
        var B = new List<Tuple<GQAPSolution, double>>();
        foreach (var v in phi) {
          int oldLocation = pi_prime[v];
          pi_prime[v] = target[v];
          var pi2 = MakeFeasible(random, pi_prime, v, nonFix, demands, capacities);
          pi_prime[v] = oldLocation;

          var currentEval = problemInstance.Evaluate(pi2);

          if (currentEval.ExcessDemand <= 0.0) {
            var quality = problemInstance.ToSingleObjective(currentEval);

            var solution = new GQAPSolution(pi2, currentEval);
            if (B.Count >= candidateSizeFactor * phi.Count) {
              int mostSimilarIndex = -1;
              double mostSimilarValue = 1.0;
              for (int i = 0; i < B.Count; i++) {
                if (quality < B[i].Item2) {
                  var similarity = HammingSimilarityCalculator.CalculateSimilarity(solution.Assignment, B[i].Item1.Assignment);
                  if (similarity == 1.0) {
                    mostSimilarIndex = -1;
                    break;
                  } else if (similarity < mostSimilarValue) {
                    mostSimilarValue = similarity;
                    mostSimilarIndex = i;
                  }
                }
              }
              if (mostSimilarIndex >= 0)
                B[mostSimilarIndex] = Tuple.Create(solution, quality);
            } else {
              bool contains = false;
              foreach (var b in B) {
                if (cmp.Equals(solution.Assignment, b.Item1.Assignment)) {
                  contains = true;
                  break;
                }
              }
              if (!contains) B.Add(Tuple.Create(solution, quality));
            }
          }
        }
        if (B.Any()) {
          var pi = B.SampleProportional(random, 1, B.Select(x => 1.0 / x.Item2), false).First();
          var diff = IntegerVectorEqualityComparer.GetDifferingIndices(pi.Item1.Assignment, target);
          var I = phi.Except(diff);
          var i = I.SampleRandom(random);
          fix.Add(i);
          nonFix.Remove(i);
          pi_prime = pi.Item1.Assignment;
          if (pi.Item2 < resultQuality) {
            result = pi.Item1.Assignment;
            resultQuality = pi.Item2;
          }
        } else return result;
        phi = new HashSet<int>(IntegerVectorEqualityComparer.GetDifferingIndices(pi_prime, target));
      }

      return result;
    }

    protected override IntegerVector Cross(IRandom random, ItemArray<IntegerVector> parents,
      GQAPInstance problemInstance) {
      return Apply(random, parents, QualityParameter.ActualValue, problemInstance,
        CandidateSizeFactorParameter.Value.Value);
    }

    private static IntegerVector MakeFeasible(IRandom random, IntegerVector assignment, int equipment, HashSet<int> nonFix, DoubleArray demands, DoubleArray capacities, int maximumTries = 1000) {
      int l = assignment[equipment];
      var slack = ComputeSlack(assignment, demands, capacities);
      if (slack[l] >= 0) return (IntegerVector)assignment.Clone();

      IntegerVector result = null;
      int k = 0;
      while (k < maximumTries && slack[l] < 0) {
        result = (IntegerVector)assignment.Clone();
        do {
          var maxSlack = slack.Max();
          var T = nonFix.Where(x => x != equipment && assignment[x] == l && demands[x] <= maxSlack);
          if (T.Any()) {
            int i = T.SampleProportional(random, 1, T.Select(x => demands[x]), false).First();
            var j = Enumerable.Range(0, capacities.Length).Where(x => slack[x] >= demands[i]).SampleRandom(random);
            result[i] = j;
            slack[j] -= demands[i];
            slack[l] += demands[i];
          } else break;
        } while (slack[l] < 0);
        k++;
      }
      return result;
    }

    private static DoubleArray ComputeSlack(IntegerVector assignment, DoubleArray demands, DoubleArray capacities) {
      var slack = (DoubleArray)capacities.Clone();
      for (int i = 0; i < assignment.Length; i++) {
        slack[assignment[i]] -= demands[i];
      }
      return slack;
    }
  }
}
